# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
this module provides classes to format the native tree2tree output
"""

import sys
try:
    from xml.dom import EMPTY_NAMESPACE as NO_NS
except:
    NO_NS = None
from xmldiff.objects import A_N1, A_N2, A_DESC, xml_print, f_xpath


def get_attrs_string(attrs):
    """ extract and return a string corresponding to an attributes list """
    attr_s = []
    for attr_n, attr_v in attrs:
        attr_s.append('%s="%s" ' % (attr_n, attr_v))
    return ' '.join(attr_s)


## Formatter interface ########################################################
class AbstractFormatter:
    """
    Formatter interface
    """

    def init(self, stream=None):
        """ method called before the begining of the tree 2 tree correction """
        if stream is None:
            stream = sys.stdout
        self.edit_s = []
        self._stream = stream

    def add_action(self, action):
        """ method called when an action is added to the edit script """
        self.edit_s.append(action)

    def format_action(self, action):
        """ method called by end() to format each action in the edit script
        at least this method should be overridden
        """
        raise NotImplementedError()

    def end(self):
        """ method called at the end of the tree 2 tree correction """
        for action in self.edit_s:
            self.format_action(action)


## Internal Formatter ##########################################################

class InternalPrinter(AbstractFormatter):
    """ print actions in the internal format """

    def add_action(self, action):
        """
        See AbstractFormatter interface
        """
        if len(action) > 2 and isinstance(action[A_N2], list):
            if isinstance(action[A_N1], list):
                # swap or move node
                action[A_N1] = f_xpath(action[A_N1])
                action[A_N2] = f_xpath(action[A_N2])
        AbstractFormatter.add_action(self, action)

    def format_action(self, action):
        """
        See AbstractFormatter interface
        """
        if len(action) > 2 and isinstance(action[A_N2], list):
            self._stream.write('[%s, %s,\n' % (action[A_DESC], action[A_N1]))
            xml_print(action[A_N2], stream=self._stream)
            self._stream.write("]\n")
        elif len(action) > 2:
            self._stream.write('[%s, %s, %s]\n' % (action[A_DESC],
                                                   action[A_N1],
                                                   action[A_N2]))
        else:
            self._stream.write('[%s, %s]\n' % (action[A_DESC], action[A_N1]))



""" Provides functions for converting DOM tree or xml file in order to process
it with xmldiff functions. """
# Copyright (c) 2001 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

from xml.sax import make_parser, SAXNotRecognizedException
from xml.sax.handler import feature_namespaces, feature_external_ges
from xml.sax.handler import feature_external_pes, property_lexical_handler

from xmldiff.parser import SaxHandler


def tree_from_stream(stream,
                     norm_sp=1, ext_ges=0, ext_pes=0, include_comment=1,
                     encoding='UTF-8', html=0):
    """
    create internal tree from xml stream (open file or IOString)
    if norm_sp = 1, normalize space and new line
    """
    handler = SaxHandler(norm_sp, include_comment, encoding)
    if html:
        parser = make_parser(["xml.sax.drivers2.drv_sgmlop_html"])
    else:
        parser = make_parser()
        # do not perform Namespace processing
        parser.setFeature(feature_namespaces, 0)
    # do not include any external entities
    try:
        parser.setFeature(feature_external_ges, ext_ges)
        # xml.sax._exceptions.
    except SAXNotRecognizedException:
        print('Unable to set feature external ges')
    try:
        parser.setFeature(feature_external_pes, ext_pes)
        # xml.sax._exceptions.
    except SAXNotRecognizedException:
        print('Unable to set feature external pes')
    # add lexical handler for comments,  entities, dtd and cdata
    parser.setProperty(property_lexical_handler, handler)
    parser.setContentHandler(handler)
    parser.parse(stream)
    return handler.get_tree()




import random
import unittest
from xmldiff.difflib import lcsl
from xmldiff.difflib import lcs2
from xmldiff.difflib import lcs4


def _cmp(a, b):
    return a == b


class TestLcs2(unittest.TestCase):
    def help_test(self, seq1, seq2, res):
        seq = lcs2(seq1, seq2, _cmp)
        self.assertEqual(seq, zip(res, res))

    def test_lcs_1(self):
        self.help_test("abcdefghijkl", "bcdeghijk", "bcdeghijk")

    def test_lcs_2(self):
        self.help_test("abdefghijkl", "bcdeghijk", "bdeghijk")

    def test_lcs_3(self):
        self.help_test("abdefghijkl", "bxcydzewgzhijk", "bdeghijk")

    def test_lcs_4(self):
        self.help_test("abdefghijkl", "zzzbcdeghijk", "bdeghijk")


# def test_time_lcs2(lcs2=lcs2):
#     import time
#     t = time.clock()
#     quick_ratio('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100)
#     print 'quick ratio :', time.clock()-t
#     lcs2('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100,
#          lambda x, y: x == y)
#     print 'lcs2 :       ', time.clock()-t
#     quick_ratio('abcdefghijklmno'*100, 'zyxwvutsrqp'*100)
#     print 'quick ratio :', time.clock()-t
#     lcs2('abcdefghijklmno'*100, 'zyxwvutsrqp'*100, lambda x, y: x == y)
#     print 'lcs2 :       ', time.clock()-t
#     quick_ratio('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100)
#     print 'quick ratio :', time.clock()-t
#     lcs2('abcdefghijklmnopqrst'*100, 'abcdefghijklmnopqrst'*100,
#          lambda x, y: x == y)
#     print 'lcs2 :       ', time.clock()-t
#     quick_ratio('abcdefghijklmno'*100, 'zyxwvutsrqp'*100)
#     print 'quick ratio :', time.clock()-t
#     lcs2('abcdefghijklmno'*100, 'zyxwvutsrqp'*100, lambda x, y: x == y)
#     print 'lcs2 :       ', time.clock()-t


# def test_main_lcs2(lcs2=lcs2):
#     print "abcde - bydc"
#     print lcsl('abcde', 'bydc', lambda x, y: x == y)
#     for a in lcs2('abcde', 'bydc', lambda x, y: x == y):
#         print a
#     print "abacdge - bcdg"
#     print lcsl('abacdge', 'bcdg', lambda x, y: x == y)
#     for a in lcs2('abacdge', 'bcdg', lambda x, y: x == y):
#         print a


def randstr(lmin, lmax, alphabet):
    L = random.randint(lmin, lmax)
    S = []
    N = len(alphabet)-1
    for i in range(L):
        S.append(alphabet[random.randint(0, N)])
    return "".join(S)


def test_random_string():
    """Generate random test sequences and compare lcs2, lcs3, lcs4"""
    import xmldiff.maplookup
    lcsm = xmldiff.maplookup.lcs2

    _alpha = "abcdefghijklmnopqrstuvwxyz"
    for i in range(100):
        S1 = randstr(2, 5, _alpha)
        S2 = randstr(2, 5, _alpha)
        # print S1, S2
        R1 = lcs2(S1, S2, _cmp)
        # print "lcs2:", "".join([x[0] for x in R1])
        R2 = lcs4(S1, S2, _cmp)
        # print "lcs4", "".join([x[0] for x in R2])
        R3 = lcsm(S1, S2, _cmp)
        # print "lcsm", "".join([x[0] for x in R3])
        # print
        assert R1 == R2, (S1, S2)
        assert R1 == R3, (S1, S2)



#!/usr/bin/python
# Copyright (c) 2001-2010 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import argparse
import sys
import os
import pkg_resources

from xmldiff.fmes import FmesCorrector
from xmldiff.format import InternalPrinter
from xmldiff.input import tree_from_stream
from xmldiff.misc import process_dirs, list_print
from xmldiff.objects import node_repr, N_ISSUE
from xml.sax import SAXParseException


def process_files(file1, file2, norm_sp, verbose,
                  ext_ges, ext_pes, include_comment, encoding,
                  html):
    """
    Computes the diff between two files.
    """
    trees = []
    for fname in (file1, file2):
        with open(fname, 'r') as fhandle:
            try:
                tree = tree_from_stream(fhandle, norm_sp, ext_ges,
                                        ext_pes, include_comment,
                                        encoding, html)
            except SAXParseException as err:
                print(err)
                return -1
            trees.append(tree)

    if verbose:
        print('Source tree:\n%s' % node_repr(trees[0]))
        print('Destination tree:\n%s' % node_repr(trees[1]))
        print('Source tree has %d nodes' % trees[0][N_ISSUE])
        print('Destination tree has %d nodes' % trees[1][N_ISSUE])

    # output formatter
    formatter = InternalPrinter()
    # choose and apply tree to tree algorithm
    strategy = FmesCorrector(formatter)
    strategy.process_trees(*trees)
    return len(formatter.edit_s)


def parse_args(argv):
    package = pkg_resources.get_distribution("xmldiff")

    parser = argparse.ArgumentParser(
        description=('Tree 2 tree correction between xml documents. '
                     'Extract differences between two xml files. '
                     'It returns a set of primitives to apply on source tree '
                     'to obtain the destination tree.'))
    parser.add_argument('-V', '--version', action='version',
                        version=package.version)
    parser.add_argument('-H', '--html', action='store_true', default=False,
                        help=('input files are HTML instead of XML.'))
    parser.add_argument('-r', '--recursive', action='store_true', default=False,
                        help=('when comparing directories, recursively compare '
                              'any subdirectories found.'))
    parser.add_argument('-e', '--encoding', default='UTF-8',
                        help=('specify the encoding to use for output. '
                              'Default is UTF-8'))
    parser.add_argument('-v', '--verbose', action='store_true', default=False)
    parser.add_argument('-n', '--not-normalize-spaces', action='store_true',
                        default=False,
                        help=('do not normalize spaces and new lines in text '
                              'and comment nodes.'))
    parser.add_argument('-c', '--exclude-comments', action='store_true',
                        default=False,
                        help=('do not process comment nodes.'))
    parser.add_argument('-g', '--ext-ges', action='store_true', default=False,
                        help=('include all external general (text) entities.'))
    parser.add_argument('-p', '--ext-pes', action='store_true', default=False,
                        help=('include all external parameter entities, '
                              'including the external DTD subset.'))

    parser.add_argument('from_file_or_dir',
                        help=('in'))
    parser.add_argument('to_file_or_dir',
                        help=('out'))

    args = parser.parse_args(argv)
    return args


def run(argv=None):
    """
    Main. To be called with list of command-line arguments (if provided,
    args should not contain the executable as first item)
    """
    if argv is None:
        argv = sys.argv[1:]

    args = parse_args(argv)

    fpath1 = args.from_file_or_dir
    fpath2 = args.to_file_or_dir
    normalize_spaces = not args.not_normalize_spaces
    include_comments = not args.exclude_comments
    exit_status = 0
    # if args are directory
    if os.path.isdir(fpath1) and os.path.isdir(fpath2):
        common, deleted, added = process_dirs(fpath1, fpath2, args.recursive)

        list_print(deleted[0], 'FILE:', 'deleted')
        list_print(deleted[1], 'DIRECTORY:', 'deleted')
        list_print(added[0], 'FILE:', 'added')
        list_print(added[1], 'DIRECTORY:', 'added')
        exit_status += sum((len(deleted[0]), len(deleted[1]),
                            len(added[0]), len(added[1])))
        for filename in common[0]:
            print('-'*80)
            print('FILE: %s' % filename)
            diffs = process_files(
                os.path.join(fpath1, filename),
                os.path.join(fpath2, filename),
                normalize_spaces, args.verbose,
                args.ext_ges, args.ext_pes, include_comments,
                args.encoding, args.html)
            if diffs:
                exit_status += diffs
    # if args are files
    elif os.path.isfile(fpath1) and os.path.isfile(fpath2):
        exit_status = process_files(
            fpath1, fpath2,
            normalize_spaces, args.verbose,
            args.ext_ges, args.ext_pes, include_comments,
            args.encoding, args.html)
    else:
        exit_status = -1
        print('%s and %s are not comparable, or not directory '
              'nor regular files' % (fpath1, fpath2))
    sys.exit(exit_status)


if __name__ == '__main__':
    run()



# Copyright (c) 2000 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""
provides constantes for using node and action (list) and some functions
for these objects use

    /!\  /!\ do not call index, remove or compare two node with == since a
             node is a recursive list
"""

import sys
from xmldiff.misc import TRUE, FALSE

################ ACTIONS #######################################################

A_DESC = 0  # string describes the action
A_N1 = 1  # node on which the action applies
A_N2 = 2  # optionnal second action argument, maybe node or value


def actp(act):
    """ print an internal action (debugging purpose) """
    if len(act) > 2:
        if act[A_DESC][0] == 'm':
            print >> sys.stderr,  act[A_DESC], caract(act[A_N1])
            print >> sys.stderr, '    ', caract(act[A_N2])
            print >> sys.stderr, '    ', caract(act[-2]), act[-3], get_pos(act[-1])
        else:
            print >> sys.stderr, act[A_DESC], caract(act[A_N1]),\
                caract(act[A_N2]),\
                act[A_N2][N_VALUE]
    else:
        print >> sys.stderr, act[A_DESC], caract(act[A_N1])


################## NODES CONSTANTES ############################################

N_TYPE = 0  # node's type
N_NAME = 1  # node's label (to process xpath)
N_VALUE = 2  # node's value
N_CHILDS = 3  # nodes's childs list
N_PARENT = 4  # node's parent
N_ISSUE = 5  # node's total issue number
N_XNUM = 6  # to compute node's xpath
NSIZE = 7  # number of items in a list which represent a node

# NODE TYPES
# NT_SYST = 0 # SYSTEM node (added by parser) /!\ deprecated
NT_NODE = 1  # ELEMENT node
NT_ATTN = 2  # ATTRIBUTE NAME node
NT_ATTV = 3  # ATTRIBUTE VALUE node
NT_TEXT = 4  # TEXT node
NT_COMM = 5  # COMMENT node
NT_ROOT = 6  # root node
NODES_TYPES = ('NT', 'NN', 'AN', 'AV', 'T', 'C', 'R')  # for printing


################## OPERATIONS EDITING NODES ####################################

def link_node(parent, child):
    """ link child to his parent """
    if child:
        parent[N_CHILDS].append(child)
        child[N_PARENT] = parent


def insert_node(node, new, pos):
    """ insert child new on node at position pos (integer) """
    node[N_CHILDS].insert(pos, new)
    new[N_PARENT] = node
    i, j = 0, 1
    while i < len(node[N_CHILDS]):
        n = node[N_CHILDS][i]
        if n[N_NAME] == new[N_NAME] and n[N_TYPE] == new[N_TYPE]:
            n[N_XNUM] = j
            j += 1
        i += 1


def delete_node(node):
    """ delete a node from its tree """
    siblings = node[N_PARENT][N_CHILDS]
    i = get_pos(node)
    siblings.pop(i)
    node[N_PARENT] = None
    while i < len(siblings):
        n = siblings[i]
        if n[N_NAME] == node[N_NAME] and n[N_TYPE] == node[N_TYPE]:
            n[N_XNUM] -= 1
        i += 1


def rename_node(node, new_name):
    """ rename a node
    this is necessary for xpath
    """
    siblings = node[N_PARENT][N_CHILDS]
    pos = get_pos(node)
    xnum = 1
    for i in range(len(siblings)):
        n = siblings[i]
        if i < pos:
            if n[N_NAME] == new_name and n[N_TYPE] == node[N_TYPE]:
                xnum += 1
        elif i != pos:
            if n[N_NAME] == node[N_NAME] and n[N_TYPE] == node[N_TYPE]:
                n[N_XNUM] -= 1
            elif n[N_NAME] == new_name and n[N_TYPE] == node[N_TYPE]:
                n[N_XNUM] += 1
    node[N_NAME] = new_name
    node[N_XNUM] = xnum


################## OPERATIONS FORMATING NODES ##################################

def caract(node):
    """ return a string which represent the node """
    return '%s:%s (%s) %s %s' % (NODES_TYPES[node[N_TYPE]], node[N_VALUE],
                                 f_xpath(node), id(node), node[N_ISSUE])


def f_xpath(node, x=''):
    """ compute node's xpath """
    if node[N_NAME] != '/':
        if node[N_TYPE] == NT_ATTN:
            return f_xpath(node[N_PARENT],
                           '/%s' % node[N_NAME][:len(node[N_NAME])-4])
        if node[N_TYPE] == NT_ATTV:
            return f_xpath(node[N_PARENT])  # [N_PARENT], '/%s'%node[N_NAME])
        return f_xpath(node[N_PARENT], '/%s[%d]%s' % (
            node[N_NAME], node[N_XNUM], x))
    elif not x:
        return '/'
    return x


def node_repr(node):
    """ return a string which represents the given node """
    s = '%s\n' % caract(node)
    for child in node[N_CHILDS]:
        s = '%s%s' % (s, _indent(child, '  '))
    return s


def _indent(node, indent_str):
    s = '%s\-%s\n' % (indent_str, caract(node))
    if next_sibling(node) is not None:
        indent_str = '%s| ' % indent_str
    else:
        indent_str = '%s  ' % indent_str
    for child in node[N_CHILDS]:
        s = '%s%s' % (s, _indent(child, indent_str))
    return s


def xml_print(node, indent='', stream=None):
    """
    recursive function which write the node in an xml form without the added
    nodes
    """
    if stream is None:
        stream = sys.stdout
    _xml_print_internal_format(node, indent, stream)


def _xml_print_internal_format(node, indent, stream):
    if node[N_TYPE] == NT_NODE:
        attrs_s = ''
        i = 0
        while i < len(node[N_CHILDS]):
            n = node[N_CHILDS][i]
            if n[N_TYPE] == NT_ATTN:
                i += 1
                attrs_s = '%s %s="%s"' % (attrs_s, n[N_VALUE],
                                          n[N_CHILDS][0][N_VALUE])
            else:
                break
        if len(node[N_CHILDS]) > i:
            stream.write('%s<%s%s>\n' % (indent, node[N_VALUE], attrs_s))
            for _curr_node in node[N_CHILDS][i:]:
                _xml_print_internal_format(
                    _curr_node, indent + '  ', stream=stream)
            stream.write('%s</%s>\n' % (indent, node[N_VALUE]))
        else:
            stream.write('%s<%s%s/>\n' % (indent, node[N_VALUE], attrs_s))
    elif node[N_TYPE] == NT_ATTN:
        stream.write('%s<@%s>\n' % (indent, node[N_VALUE]))
        stream.write(node[N_CHILDS][0][N_VALUE] + '\n')
        stream.write('%s</%s>\n' % (indent, node[N_VALUE]))
    elif node[N_TYPE] == NT_COMM:
        stream.write('%s<!-- %s -->\n' % (indent, node[N_VALUE]))
    elif node[N_TYPE] == NT_TEXT:
        stream.write(node[N_VALUE] + '\n')
    else:
        stream.write('unknown node type', str(node[N_TYPE]))


def to_dom(node, doc, uri=None, prefix=None):
    """
    recursive function to convert internal tree in an xml dom tree without
    the added nodes
    """
    if node[N_TYPE] == NT_NODE:
        dom_n = doc.createElementNS(uri, '%selement' % prefix)
        dom_n.setAttributeNS(None, 'name', node[N_VALUE])
        for n in node[N_CHILDS]:
            if n[N_TYPE] == NT_ATTN:
                dom_n = doc.createElementNS(uri, '%sattribute' % prefix)
                v = unicode(n[N_CHILDS][0][N_VALUE], 'UTF-8')
                dom_n.setAttributeNS(None, 'name', n[N_VALUE])
                dom_n.appendChild(doc.createTextNode(v))
            else:
                dom_n.appendChild(to_dom(n, doc, uri))
    elif node[N_TYPE] == NT_ATTN:
        dom_n = doc.createElementNS(uri, '%sattribute' % prefix)
        dom_n.setAttributeNS(None, 'name', node[N_VALUE])
        v = unicode(node[N_CHILDS][0][N_VALUE], 'UTF-8')
        dom_n.appendChild(doc.createTextNode(v))
    elif node[N_TYPE] == NT_COMM:
        dom_n = doc.createElementNS(uri, '%scomment' % prefix)
        v = unicode(node[N_VALUE], 'UTF-8')
        dom_n.appendChild(doc.createTextNode(v))
    elif node[N_TYPE] == NT_TEXT:
        dom_n = doc.createElementNS(uri, '%stext' % prefix)
        v = unicode(node[N_VALUE], 'UTF-8')
        dom_n.appendChild(doc.createTextNode(v))
    return dom_n


################## OPERATIONS GIVING INFOS ON NODES ############################

def get_pos(node):
    """ return the index of a node in its parent's children list

    /!\  /!\ do not call index, remove or compare two node with == since a
             node is a recursive list
    """
    try:
        childs = node[N_PARENT][N_CHILDS]
        for i, child in enumerate(childs):
            if child is node:
                return i
    except TypeError:
        return -1
    except ValueError:
        return -1


def nb_childs(node):
    """ return the number of childs (without attribute childs) of the given node
    """
    return len(filter(lambda n: n[N_CHILDS][0][N_TYPE] != NT_ATTN,
                      node[N_CHILDS]))


def nb_attrs(node):
    """ return the number of attributes of the given node """
    for i, child in enumerate(node[N_CHILDS]):
        if child[N_TYPE] != NT_ATTN:
            break
    else:
        try:
            i += 1
        except UnboundLocalError:
            i = 0
    return i


################## MISCELLANEOUS OPERATIONS ON NODES ###########################

def next_sibling(node):
    """ return the node's right sibling """
    if node[N_PARENT] is None:
        return None
    myindex = get_pos(node)
    if len(node[N_PARENT][N_CHILDS]) > myindex+1:
        return node[N_PARENT][N_CHILDS][myindex+1]
    return None


def previous_sibling(node):
    """ return the node's left sibling """
    myindex = get_pos(node)
    if node[N_PARENT] and myindex > 0:
        return node[N_PARENT][N_CHILDS][myindex-1]
    return None


def get_ancestors(node, l):
    """ append to l all the ancestors from node """
    while node[N_PARENT]:
        l.append(node)
        node = node[N_PARENT]
    return l


def get_labels(tree, labels, leaf_labels):
    """
    Chain all nodes with a given label l in tree T together, from left to right,
    by filling dictionnaries labels and leaf_labels (for leaf nodes).

    Label are keys pointing to a list of nodes with this type.
    Node x occurs after y in the list if x appears before y in the in-order
    traversal of T.
    /!\  /!\
    since this isn't binary tree, post order traversal (?)
    """
    if tree and tree[N_CHILDS]:
        for node in tree[N_CHILDS]:
            get_labels(node,  labels, leaf_labels)
        labels.setdefault(NODES_TYPES[tree[N_TYPE]], []).append(tree)
    elif tree:
        leaf_labels.setdefault(NODES_TYPES[tree[N_TYPE]], []).append(tree)


def make_bfo_list(tree):
    """ create a list with tree nodes in breadth first order """
    l, queue = [], []
    if tree:
        l.append(tree)
        if tree[N_CHILDS]:
            node = tree[N_CHILDS][0]
            while node:
                l.append(node)
                if node[N_CHILDS]:
                    queue.append(node)
                node = next_sibling(node)
                if not node and queue:
                    node = queue.pop(0)[N_CHILDS][0]
    return l


def make_bfo_list(tree):
    """ create a list with tree nodes in breadth first order """
    queue = [tree]
    lst = [tree]
    while queue:
        node = queue.pop(0)
        lst.extend(node[N_CHILDS])
        queue.extend([n for n in node[N_CHILDS] if n[N_CHILDS]])
    return lst

# no more used
# def make_po_list(tree):
##     """ create a list with tree nodes in post order """
##     l, stack, poped = [], [], 0
# if tree:
# if tree[N_CHILDS]:
##             node = tree[N_CHILDS][0]
# while node:
# if node[N_CHILDS] and not poped:
# stack.append(node)
##                     node = node[N_CHILDS][0]
# else:
# l.append(node)
##                     node = next_sibling(node)
##                 poped = 0
# if not node and stack:
##                     node = stack.pop()
##                     poped = 1
# l.append(tree)
# return l
# def make_preo_list(tree):
##     """ create a list with tree nodes in pre order """
##     l, stack, poped = [], [], 0
# if tree:
# l.append(tree)
# if tree[N_CHILDS]:
##             node = tree[N_CHILDS][0]
# xl.append(node)
# while node:
# if node[N_CHILDS] and not poped:
# stack.append(node)
##                     node = node[N_CHILDS][0]
# else:
##                     node = next_sibling(node)
# l.append(node)
##                 poped = 0
# if not node and stack:
##                     node = stack.pop()
##                     poped = 1
# return l
# def get_leafs(tree, l):
##     """ return a list with all leaf nodes from left to right """
# if tree and tree[N_CHILDS]:
##         node = tree[N_CHILDS][0]
# while node:
##             get_leafs(node, l)
##             node = next_sibling(node)
# elif tree:
# l.append(tree)
# def get_issue(node, l):
##     """ append to l all the descendants from node """
# for child in node[N_CHILDS]:
# l.append(child)
# if child[N_CHILDS]:
##             get_issue(child, l)
# def contains(ancestor, node):
##     """ return true if node is descendent of ancestor """
# if node is None:
# return FALSE
# if ancestor is node:
# return TRUE
# return contains(ancestor, node[N_PARENT])



"""
xmldiff non regression test
"""
from os.path import join, basename, dirname
import sys
import six
import pytest
import glob

from xmldiff import main

HERE = dirname(__file__)


def get_output(options, expected):
    backup = sys.stdout

    # capture stdout
    sys.stdout = six.StringIO()
    try:
        main.run(options)
    except SystemExit:
        pass
    finally:
        output = sys.stdout.getvalue().strip()
        sys.stdout.close()
        sys.stdout = backup

    return output


def test_recursive():
    options = ['-r', join(HERE, 'data', 'dir1'), join(HERE, 'data', 'dir2')]
    expected = """--------------------------------------------------------------------------------
FILE: onlyindir1.xml deleted
--------------------------------------------------------------------------------
FILE: onlyindir2.xml added
--------------------------------------------------------------------------------
FILE: inbothdir.xml"""
    data = get_output(options, expected)
    assert data == expected, '%s:\n%r != %r' % (options, data, expected)


def test_broken():
    options = ['-r', join(HERE, 'data', 'broken', 'broken.xml'),
               join(HERE, 'data', 'broken', 'broken.xml')]
    expected = "xmldiff/tests/data/broken/broken.xml:11:4: mismatched tag"
    data = get_output(options, expected)
    assert expected in data


def make_tests():
    """generate tests classes from test info

    return the list of generated test classes
    """
    tests_files = glob.glob(join(HERE, 'data', '*.xml')) + \
        glob.glob(join(HERE, 'data', '*_result'))
    tests = {}
    # regroup test files
    for filename in tests_files:
        base = basename(filename)
        name = base[:6]
        filetype = base[7:]
        if filetype == '1.xml':
            tests.setdefault(name, {})['old'] = filename
        elif filetype == '2.xml':
            tests.setdefault(name, {})['new'] = filename
        else:
            tests.setdefault(name, {})['result'] = filename

    result = []
    for t_name, t_dict in tests.items():
        # quick check whether input, output, result is there
        old = t_dict['old']
        new = t_dict['new']
        res_data = t_dict['result']

    return tests.values()


@pytest.fixture(params=make_tests())
def fnames(request):
    return request.param


def test_known(fnames):
    old = fnames['old']
    new = fnames['new']
    res_file = fnames['result']
    f = open(res_file)
    expected = f.read().strip()
    f.close()
    options = [old, new]
    data = get_output(options, expected)
    assert data == expected, '%s:\n%r != %r' % (options, data, expected)



"""
longest common subsequence algorithm

the algorithm is describe in "An O(ND) Difference Algorithm and its Variation"
by Eugene W. MYERS

As opposed to the algorithm in difflib.py, this one doesn't require hashable
elements 
"""


def lcs2(X, Y, equal):
    """
    apply the greedy lcs/ses algorithm between X and Y sequence
    (should be any Python's sequence)
    equal is a function to compare X and Y which must return 0 if
    X and Y are different, 1 if they are identical
    return a list of matched pairs in tuplesthe greedy lcs/ses algorithm
    """
    N, M = len(X), len(Y)
    if not X or not Y:
        return []
    max = N + M
    v = [0 for i in xrange(2*max+1)]
    common = [[] for i in xrange(2*max+1)]
    for D in xrange(max+1):
        for k in xrange(-D, D+1, 2):
            if k == -D or k != D and v[k-1] < v[k+1]:
                x = v[k+1]
                common[k] = common[k+1][:]
            else:
                x = v[k-1] + 1
                common[k] = common[k-1][:]

            y = x - k
            while x < N and y < M and equal(X[x], Y[y]):
                common[k].append((x, y))
                x += 1
                y += 1

            v[k] = x
            if x >= N and y >= M:
                return [(X[x], Y[y]) for x, y in common[k]]


def lcs4(X, Y, equal):
    """
    apply the greedy lcs/ses algorithm between X and Y sequence
    (should be any Python's sequence)
    equal is a function to compare X and Y which must return 0 if
    X and Y are different, 1 if they are identical
    return a list of matched pairs in tuplesthe greedy lcs/ses algorithm
    """
    N, M = len(X), len(Y)
    if not X or not Y:
        return []
    max = N + M
    v = [0 for i in xrange(2*max+1)]
    vl = [v]
    for D in xrange(max+1):
        for k in xrange(-D, D+1, 2):
            if k == -D or k != D and v[k-1] < v[k+1]:
                x = v[k+1]
            else:
                x = v[k-1] + 1

            y = x - k
            while x < N and y < M and equal(X[x], Y[y]):
                x += 1
                y += 1

            v[k] = x
            if x >= N and y >= M:
                # reconstruction du chemin
                vl.append(v)
                vl_saved = vl[:]
                path = []
                k = N-M

                while vl:
                    oldv = vl.pop(-1)
                    oldk = k
                    if k == -D or k != D and oldv[k-1] < oldv[k+1]:
                        xs = oldv[k+1]
                        k = k + 1
                    else:
                        xs = oldv[k-1]+1
                        k = k - 1
                    # print "-> x=%d y=%d v=%r ok=%d k=%d xs=%d D=%d" % (x,y,oldv,oldk,k,xs,D)
                    while x > xs:
                        x -= 1
                        y -= 1
                        # print "(%d,%d)" % (x,y)
                        path.append((X[x], Y[y]))
                    D -= 1
                    x = oldv[k]
                    y = x - k
                    # print "<- x=%d y=%d v=%r ok=%d k=%d xs=%d D=%d" % (x,y,oldv,oldk,k,xs,D)
                # print x,y,deltax,deltay,oldv, oldk, k
                path.reverse()
                return path  # , vl_saved
        vl.append(v[:])


def lcs3(X, Y, equal):
    N = len(X)+1
    M = len(Y)+1
    if not X or not Y:
        return []
    # D(i,j) is the length of longest subsequence for X[:i], Y[:j]
    pre = [0]*M
    row = [0]*M
    B = [[0]*M for i in xrange(N)]
    for i in xrange(1, N):
        for j in xrange(1, M):
            if equal(X[i-1], Y[j-1]):
                row[j] = pre[j-1] + 1
                B[i][j] = 2  # move back (-1,-1)
            elif pre[j] >= row[j-1]:
                row[j] = pre[j]
                B[i][j] = 1  # move back (0,-1)
            else:
                row[j] = row[j-1]
                B[i][j] = 0  # move back (-1,0)
        pre, row = row, pre
    i = N - 1
    j = M - 1
    L = []
    while i >= 0 and j >= 0:
        d = B[i][j]
        # print i,j,d
        if d == 0:
            j -= 1
        elif d == 1:
            i -= 1
        else:
            i -= 1
            j -= 1
            L.append((X[i], Y[j]))
    L.reverse()
    #from pprint import pprint
    # pprint(D)
    # pprint(B)
    return L


try:
    import xmldiff.maplookup
    lcs2 = xmldiff.maplookup.lcs2
    #lcs2 = lcs4
except:
    pass


def lcsl(X, Y, equal):
    """return the length of the result sent by lcs2"""
    return len(lcs2(X, Y, equal))


def quick_ratio(a, b):
    """
    optimized version of the standard difflib.py quick_ration
    (without junk and class)
    Return an upper bound on ratio() relatively quickly.
    """
    # viewing a and b as multisets, set matches to the cardinality
    # of their intersection; this counts the number of matches
    # without regard to order, so is clearly an upper bound
    if not a and not b:
        return 1
    fullbcount = {}
    for elt in b:
        fullbcount[elt] = fullbcount.get(elt, 0) + 1
    # avail[x] is the number of times x appears in 'b' less the
    # number of times we've seen it in 'a' so far ... kinda
    avail = {}
    matches = 0
    for elt in a:
        if elt in avail:
            numb = avail[elt]
        else:
            numb = fullbcount.get(elt, 0)
        avail[elt] = numb - 1
        if numb > 0:
            matches = matches + 1
    return 2.0 * matches / (len(a) + len(b))



#!/usr/bin/env python
# pylint: disable-msg=W0142, W0403,W0404, W0613,W0622,W0622, W0704, R0904
#
# Copyright (c) 2003-2010 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
"""Setup
"""
import os
from setuptools import setup, find_packages


def read(*rnames):
    with open(os.path.join(os.path.dirname(__file__), *rnames), 'rb') as f:
        return f.read().decode('utf-8')


TESTS_REQUIRE = [
    'coverage',
    'mock',
   ]

try:
    from distutils.core import Extension
    ext_modules = [Extension('xmldiff.maplookup',
                             ['extensions/maplookup.c'])]
except:
    ext_modules = []


setup(
    name='xmldiff',
    version='1.0.0a1',
    author="Shoobx Team",
    author_email="dev@shoobx.com",
    url='https://github.com/Shoobx/xmldiff',
    description=('Tree 2 tree correction between xml documents. '
                 'Extract differences between two xml files. '
                 'It returns a set of primitives to apply on source tree '
                 'to obtain the destination tree.'),
    long_description=(
        read('src', 'xmldiff', 'README.txt') +
        '\n\n' +
        read('CHANGES.txt')
    ),
    license='LGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Framework :: ZODB',
        'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
        'Natural Language :: English',
        'Operating System :: OS Independent'],
    packages=find_packages('src'),
    package_dir={'': 'src'},
    extras_require=dict(
        test=TESTS_REQUIRE,
    ),
    install_requires=[
        'future',
        'six',
        'setuptools',
    ],
    include_package_data=True,
    zip_safe=False,
    entry_points='''
    [console_scripts]
    xmldiff = xmldiff.main:run
    ''',
    ext_modules=ext_modules,
    tests_require=TESTS_REQUIRE,
)
